from typing import List, Union

from quple.circuit.quantum_circuit import QuantumCircuit
from quple.components.parametrisation import UnresolvedParameters
from sympy import Symbol

class ParametrisedCircuit(QuantumCircuit):
    def __init__(self, n_qubit:int, 
                 parameters: Union[str, List[str], Symbol, List[Symbol]]=None) -> None:
        super().__init__(n_qubit)
        self._params = UnresolvedParameters(parameters)
        
    @property
    def params(self):
        return self._params
    
        
        


from __future__ import annotations
from abc import ABC, abstractmethod

from typing import (Any, Callable, cast, Dict, FrozenSet, Iterable, Iterator,
                    List, Optional, overload, Sequence, Set, Tuple, Type,
                    TYPE_CHECKING, TypeVar, Union)
import re
import numpy as np 

import sympy

import cirq   
from cirq import  devices, GridQubit 

from quple.circuit.quantum_register import QuantumRegister


# InsertStrategy.EARLIEST, InsertStrategy.NEW, InsertStrategy.INLINE and InsertStrategy.NEW_THEN_INLINE

class QuantumCircuit(cirq.Circuit):
    def __init__(self, n_qubit:int, name:str=None) -> None:
        super().__init__()
        self._name = 'qc' or name
        self._n_qubit = n_qubit
        self._qubits = QuantumRegister(n_qubit)
        self._blocks = {} # circuit blocks

    @staticmethod
    def _parse_qubit_operation(qubit_ops:Dict, target_qubits):
        resolved_operation = None
        for op in qubit_ops:
            pass

    @property
    def name(self) -> str:
        return self._name
    
    @property
    def n_qubit(self) -> int:
        return self._n_qubit

    @property
    def circuit(self):
        return self

    
    @property
    def qubits(self):
        return self._qubits

    @property
    def diagram(self):
        print(self)
        

    def apply_gate_operation(self, operation:cirq.Gate, qubit_expr):
        qubit_sequence = self._qubits.get(qubit_expr)
        strategy = cirq.circuits.InsertStrategy.NEW_THEN_INLINE
        if all(isinstance(qubits, GridQubit) for qubits in qubit_sequence):
            self.append([operation(qubit) for qubit in qubit_sequence], strategy=strategy)
        elif all(isinstance(qubits, tuple) for qubits in qubit_sequence):
            self.append([operation(*qubits) for qubits in qubit_sequence], strategy=strategy)
        else:
            raise ValueError('Inconsistent qubit representation: {}'.format(operation, qubit_sequence))
            
    def get_param_resolver(self, param_values: Dict):
        return cirq.ParamResolver(param_values)
    
    def _test01(self, repetitions=500):
        simulator.run(self, param_resolver, repetitions=repetitions)

    def S(self, qubit_expr):
        '''
        Clifford-S gate
        '''
        self.apply_gate_operation(cirq.ops.S, qubit_expr)

    def T(self, qubit_expr):
        '''
        Non-Clifford-T gate
        '''
        self.apply_gate_operation(cirq.ops.T, qubit_expr)


    def H(self, qubit_expr):
        '''
        Hadamard gate
        '''
        self.apply_gate_operation(cirq.ops.H, qubit_expr)

    def R(self, axis:str, theta:Union[int, float], qubit_expr):
        if axis.lower() == 'x':
            self.Rx(theta, qubit_expr)
        elif axis.lower() == 'y':
            self.Ry(theta, qubit_expr)
        elif axis.lower() == 'z':
            self.Rz(theta, qubit_expr)
        else:
            raise ValueError('Invalid rotation axis: {}'.format(axis))
    
    def Rx(self, theta:Union[int, float], qubit_expr):
        self.apply_gate_operation(cirq.ops.rx(theta), qubit_expr)

    def Ry(self, theta:Union[int, float], qubit_expr):
        self.apply_gate_operation(cirq.ops.ry(theta), qubit_expr)

    def Rz(self, theta:Union[int, float], qubit_expr):
        self.apply_gate_operation(cirq.ops.rz(theta), qubit_expr)                
    
    def SWAP(self, qubit_expr):
        '''
        SWAP gate
        '''
        self.apply_gate_operation(cirq.ops.SWAP, qubit_expr)

    def apply_sequential_gate_operations(self, qubit_ops:list):
        for qop in qubit_ops:
            operation = QuantumCircuit._parse_qubit_operation(qop, self._qubits)
            self.append(operation)
            
    def PauliX(self, qubit_expr):
        self.apply_gate_operation(cirq.X, qubit_expr)
        
    def PauliY(self, qubit_expr):
        self.apply_gate_operation(cirq.Y, qubit_expr)

    def PauliZ(self, qubit_expr):
        self.apply_gate_operation(cirq.Z, qubit_expr)
        

    def CNOT(self, qubit_expr):
        '''
        Controlled NOT gate
        '''
        self.apply_gate_operation(cirq.ops.CNOT, qubit_expr)
    
    def Toffoli(self, qubit_expr):
        self.apply_gate_operation(cirq.ops.TOFFOLI, qubit_expr)

    def PhaseShift(self, phi, qubit_expr):
        self.apply_gate_operation(cirq.ZPowGate(exponent=phi / np.pi), qubit_expr)
    
    def u1(self, theta:Union[int, float], qubit_expr):
        '''
        qiskit implementation: u1 = RZ gate
        named u1 for hardware reason
        '''
        return self.Rz(theta, qubit_expr)
            
    def CX(self, qubit_expr):
        '''
        Same as CNOT gate
        '''
        self.apply_gate_operation(cirq.ops.CX, qubit_expr)
  
    def CZ(self, qubit_expr):
        '''
        Controlled Z gate
        '''
        self.apply_gate_operation(cirq.ops.CZ, qubit_expr)
        
    def clear(self) -> None:
        '''
        clear all moments
        '''
        self._moments = []
        
    @property
    def blocks(self):
        return self._blocks
    
    def new_block(self, label:str, ignore_redef:bool=True) -> None:
        if (not ignore_redef) and (label in self._blocks):
            raise RuntimeError('circuit block with label {} already exist'.format(label))
        self._blocks[label] = QuantumCircuit(self._n_qubit, name=label)
        return self._blocks[label]
            
    
    def append_block(self, label:Union[str, QuantumCircuit]) -> None:
        if isinstance(label, str):
            self += self.blocks[label]
        elif isinstance(label, QuantumCircuit):
            self += label
        else:
            raise ValueError('unknown circuit block of type {}'.format(label))

    def google_device_layouts(self, name):
        if name == 'Syncamore':
            return cirq.google.Syncamore
        elif name == 'Sycamore23':
            return cirq.google.Sycamore23
        elif name == 'Bristlecone':
            return cirq.google.Bristlecone
        elif name == 'Foxtail':
            return cirq.google.Foxtail
    '''
    def get_entangled_qubits(self) -> Set[Int]:
        entangled_qubits = set()
    '''




from functools import reduce
import numpy as np

from quple import QuantumCircuit

from .feature_map import FeatureMap 


class PauliExpansion(EncodingCircuit):
	def __init__(self, feature_dimension: int,
                 depth: int=2, paulis:List[str] = ['Z', 'ZZ']):
		super().__init__(feature_dimension, depth)
        self._paulis = paulis
        

	def construct_circuit(self, inverse=False):
        self.clear()
        n_qubit = self._n_qibit
        # build the Hadamard layer
        Hadamard_layer = self.new_block('Hadamard_layer')
        Hadamard_layer.H(range(n_qubit))
        self.append_block(Hadamard_layer)
        
        # build the entanglement layer
		
        
def self_product(x: np.ndarray) -> float:
    """
    Define a function map from R^n to R.

    Args:
        x: data

    Returns:
        float: the mapped value
    """
    coeff = x[0] if len(x) == 1 else reduce(lambda m, n: m * n, np.pi - x)  
    return coeff


'''

#reverse order of Pauli operations:
    pauli = pauli[::-1]
    

for pauli in paulis:
    mapped_params = ...
    puli_block = pauli_evolution(pauli, mapped_params)
    
    
pauli_string = pauli_string[::-1]

# trim the pauli string if identities are included
trimmed = []
indices = []
for i, pauli in enumerate(pauli_string):
    if pauli != 'I':
        trimmed += [pauli]
        indices += [i]

evo = QuantumCircuit(len(pauli_string))

if len(trimmed) == 0:
    return evo

def basis_change(circuit, inverse=False):
    for i, pauli in enumerate(pauli_string):
        if pauli == 'X':
            circuit.h(i)
        elif pauli == 'Y':
            circuit.rx(-np.pi / 2 if inverse else np.pi / 2, i)

def cx_chain(circuit, inverse=False):
    num_cx = len(indices) - 1
    for i in reversed(range(num_cx)) if inverse else range(num_cx):
        circuit.cx(indices[i], indices[i + 1])

basis_change(evo)
cx_chain(evo)
evo.u1(2.0 * time, indices[-1])
cx_chain(evo, inverse=True)
basis_change(evo, inverse=True)


get_entangler_map = list(combinations(list(range(num_circuit_qubits)), num_block_qubits))

build_entanglement_layer = 

#i : repetition
#j : the j-th block ['Z','ZZ'] = 2 blocks
#indices: the entanglement indices
for j, block in enumerate(self.entanglement_blocks):
    # create a new layer and get the entangler map for this block
    layer = QuantumCircuit(*self.qregs)
    entangler_map = self.get_entangler_map(i, j, block.num_qubits)

    # apply the operations in the layer
    for indices in entangler_map:
        parametrized_block = self._parametrize_block(block, param_iter, i, j, indices)
        layer.compose(parametrized_block, indices, inplace=True)

    # add the layer to the circuit
    self += layer


def _parametrize_block(self, block, param_iter=None, rep_num=None, block_num=None, indices=None,
                       params=None):
    """Convert ``block`` to a circuit of correct width and parameterized using the iterator."""
    if self._overwrite_block_parameters:
        # check if special parameters should be used
        # pylint: disable=assignment-from-none
        if params is None:
            params = self._parameter_generator(rep_num, block_num, indices)
        if params is None:
            params = [next(param_iter) for _ in range(len(get_parameters(block)))]

        update = dict(zip(block.parameters, params))
        return block.assign_parameters(update)

    return block.copy()
    
def compose(self, other, qubits=None, clbits=None, front=False, inplace=False):
    """Compose circuit with ``other`` circuit or instruction, optionally permuting wires.

    ``other`` can be narrower or of equal width to ``self``.

    Args:
        other (qiskit.circuit.Instruction or QuantumCircuit or BaseOperator):
            (sub)circuit to compose onto self.
        qubits (list[Qubit|int]): qubits of self to compose onto.
        clbits (list[Clbit|int]): clbits of self to compose onto.
        front (bool): If True, front composition will be performed (not implemented yet).
        inplace (bool): If True, modify the object. Otherwise return composed circuit.

    Returns:
        QuantumCircuit: the composed circuit (returns None if inplace==True).

    Raises:
        CircuitError: if composing on the front.
        QiskitError: if ``other`` is wider or there are duplicate edge mappings.

    Examples:

        >>> lhs.compose(rhs, qubits=[3, 2], inplace=True)

        .. parsed-literal::

                        ┌───┐                   ┌─────┐                ┌───┐
            lqr_1_0: ───┤ H ├───    rqr_0: ──■──┤ Tdg ├    lqr_1_0: ───┤ H ├───────────────
                        ├───┤              ┌─┴─┐└─────┘                ├───┤
            lqr_1_1: ───┤ X ├───    rqr_1: ┤ X ├───────    lqr_1_1: ───┤ X ├───────────────
                     ┌──┴───┴──┐           └───┘                    ┌──┴───┴──┐┌───┐
            lqr_1_2: ┤ U1(0.1) ├  +                     =  lqr_1_2: ┤ U1(0.1) ├┤ X ├───────
                     └─────────┘                                    └─────────┘└─┬─┘┌─────┐
            lqr_2_0: ─────■─────                           lqr_2_0: ─────■───────■──┤ Tdg ├
                        ┌─┴─┐                                          ┌─┴─┐        └─────┘
            lqr_2_1: ───┤ X ├───                           lqr_2_1: ───┤ X ├───────────────
                        └───┘                                          └───┘
            lcr_0: 0 ═══════════                           lcr_0: 0 ═══════════════════════

            lcr_1: 0 ═══════════                           lcr_1: 0 ═══════════════════════

    """
    if front:
        raise CircuitError("Front composition of QuantumCircuit not supported yet.")

    if isinstance(other, QuantumCircuit):
        from qiskit.converters.circuit_to_dag import circuit_to_dag
        from qiskit.converters.dag_to_circuit import dag_to_circuit

        dag_self = circuit_to_dag(self)
        dag_other = circuit_to_dag(other)
        dag_self.compose(dag_other, qubits=qubits, clbits=clbits, front=front)
        composed_circuit = dag_to_circuit(dag_self)
        if inplace:  # FIXME: this is just a hack for inplace to work. Still copies.
            self.__dict__.update(composed_circuit.__dict__)
            return None
        else:
            return composed_circuit

    else:  # fall back to append which accepts Instruction and BaseOperator
        if inplace:
            self.append(other, qargs=qubits, cargs=clbits)
            return None
        else:
            new_circuit = self.copy()
            new_circuit.append(other, qargs=qubits, cargs=clbits)
            return new_circuit    

'''


from typing import (Any, Callable, cast, Dict, FrozenSet, Iterable, Iterator,
                    List, Optional, overload, Sequence, Set, Tuple, Type,
                    TYPE_CHECKING, TypeVar, Union)    

import re
import sympy

class UnresolvedParameters():
    def __init__(self, params: Union[str, List[str], List[sympy.Symbol], None]=None):
        self._params = {}
        if params:
            self.append(params)
    
    @property
    def params(self) -> Dict:
        return self._params
    
    def remove_list(self, symbol:str):
        expr = '{}\[\d+\]'.format(symbol)
        matched = [ param for param in self._params if re.match(expr, param)]
        for param in matched:
            self._params.pop(param, None)
    
    def __getitem__(self, key):
        return self._params[key]
    
    def __len__(self):
        return len(self._params)
    
    def append_list(self, symbol:str, length:int):
        expanded_symbols = ['{}[{}]'.format(symbol, i) for i in range(length)]
        self.append(expanded_symbols)
        
    def keys(self):
        return self._params.keys()
    
    def values(self):
        return self._params.values()
    
    def append(self, params: Union[str, List[str], sympy.Symbol, List[sympy.Symbol]] ):
        if isinstance(params, list):
            if all(isinstance(param, str) for param in params):
                self._params.update(UnresolvedParameters._create_params(params))
            elif all(isinstance(param, sympy.Symbol) for param in params):
                self._params.update({param.name: param for param in params})
            else:
                raise ValueError('inconsistent list of parameters passed: {}'.format(params))
        elif isinstance(params, str):
            self.append([params])
        elif isinstance(params, sympy.Symbol):
            self._params[params.name] = params
        else:
            raise TypeError('unsupported parameter type : {}'.format(type(params)))
    
    
    @staticmethod
    def _create_params(params:List[str]) -> Dict:
        return {param: sympy.Symbol(param) for param in params}
        


from abc import ABC, abstractmethod

from quple.circuit.parametrised_circuit import ParametrisedCircuit


class EncodingCircuit(ParametrisedCircuit, ABC):

	def __init__(self, feature_dimenion:int, depth:int=1):
        super().__init__(feature_dimension)
		self._feature_dimension = feature_dimension
		self._depth = depth
        self.params.append_list('x', feature_dimension)

	@abstractmethod
	def construct_circuit(self, inverse=False):
		raise NotImplementedError()


	@property
	def feature_dimension(self):
		return self._feature_dimension
    
	@property
	def depth(self):
		return self._depth


from typing import List, Sequence, Union
from itertools import combinations

from quple.circuit.quantum_register import QuantumRegister

'''
if n_block_qubit > n_qubit:
    raise ValueError('The number of block qubits cannot exceed the number of '
                     'qubits in the circuit.')
'''


class EntanglerMap():
    def __init__(self, qr:Union[int, QuantumRegister], n_ent:int=2) -> None:
        '''
        qr: quantum register
        n_ent: number of entangled qubits in each block
        '''
        if isinstance(qr, int):
            self._qr = QuantumRegister(qr)
        elif isinstance(qr, QuantumRegister):
            self._qr = qr
        
        self._n_ent = n_ent
        
        if self.n_ent > self.n_qubit:
            raise ValueError('The number of entangled qubits in a block'
            'cannot exceed the total number of qubits in the circuit.')
            
            
    @property
    def n_qubit(self):
        '''
        number of circuit qubits
        '''
        return self._qr.n_qubit
    
    @property
    def n_ent(self):
        '''
        degree of entanglement
        '''
        return self._n_ent
    
    
    @property
    def qr(self):
        '''
        quantum registers involved in entangler map
        '''
        return self._qr
    
    @property
    def n_block_qubit(self):
        return self._n_block_qubit

    def full_entanglement(self) -> List[Sequence[int]]:
        n, m = self.n_qubit, self.n_ent
        return list(combinations(list(range(n)), m))

    def linear_entanglement(self) -> List[Sequence[int]]:
        n, m = self.n_qubit, self.n_ent
        return [tuple(range(i, i + m)) for i in range(n - m + 1)]

    def circular_entanglement(self) -> List[Sequence[int]]:
        n, m = self.n_qubit, self.n_ent
        return [tuple(range(n - m + 1, n)) + (0,)] + self.linear_entanglement()

    def get_entangler_map(self, method:str='full', qubit_repr:bool=False) -> List[Sequence[int]]:
        entangle_method = {
            'full': self.full_entanglement,
            'linear': self.linear_entanglement,
            'circular': self.circular_entanglement
        }
        entangler_map = entangle_method[method]()
        if qubit_repr:
            entangler_map = self._qr._parse_qubit_expression(entangler_map, self.qr.qubits)
        return entangler_map
